# signature { license=0, seed=457995617, key="8743f8ba7b391153021bf1af2774b005" }
-----------------------------------------------------------------------------








local base = _G
local string = require("string")
local math = require("math")
local socket = require("socket.core")
local _M = socket
_M._VERSION = "LuaSocket 2.0.5"



function _M.connect(address, port, laddress, lport)
	local sock, err = socket.tcp()
	if not sock then return nil, err end
	if laddress then
		local res, err = sock:bind(laddress, lport, -1)
		if not res then return nil, err end
	end
	local res, err = sock:connect(address, port)
	if not res then return nil, err end
	return sock
end

function _M.bind(host, port, backlog)
	local sock, err = socket.tcp()
	if not sock then return nil, err end
	sock:setoption("reuseaddr", true)
	local res, err = sock:bind(host, port)
	if not res then return nil, err end
	res, err = sock:listen(backlog)
	if not res then return nil, err end
	return sock
end

_M.try = _M.newtry()

function _M.choose(table)
	return function(name, opt1, opt2)
		if base.type(name) ~= "string" then
			name, opt1, opt2 = "default", name, opt1
		end
		local f = table[name or "nil"]
		if not f then base.error("unknown key (".. base.tostring(name) ..")", 3)
		else return f(opt1, opt2) end
	end
end





_M.sourcet = {}
_M.sinkt = {}

_M.BLOCKSIZE = 2048

_M.sinkt["close-when-done"] = function(sock)
	return base.setmetatable({
		getfd = function() return sock:getfd() end,
		dirty = function() return sock:dirty() end
	}, {
		__call = function(self, chunk, err)
			if not chunk then
				sock:close()
				return 1
			else return sock:send(chunk) end
		end
	})
end

_M.sinkt["keep-open"] = function(sock)
	return base.setmetatable({
		getfd = function() return sock:getfd() end,
		dirty = function() return sock:dirty() end
	}, {
		__call = function(self, chunk, err)
			if chunk then return sock:send(chunk)
			else return 1 end
		end
	})
end

_M.sinkt["default"] = _M.sinkt["keep-open"]

_M.sink = _M.choose(_M.sinkt)

_M.sourcet["by-length"] = function(sock, length)
	return base.setmetatable({
		getfd = function() return sock:getfd() end,
		dirty = function() return sock:dirty() end
	}, {
		__call = function()
			if length <= 0 then return nil end
			local size = math.min(socket.BLOCKSIZE, length)
			local chunk, err = sock:receive(size)
			if err then return nil, err end
			length = length - string.len(chunk)
			return chunk
		end
	})
end

_M.sourcet["until-closed"] = function(sock)
	local done
	return base.setmetatable({
		getfd = function() return sock:getfd() end,
		dirty = function() return sock:dirty() end
	}, {
		__call = function()
			if done then return nil end
			local chunk, err, partial = sock:receive(socket.BLOCKSIZE)
			if not err then return chunk
			elseif err == "closed" then
				sock:close()
				done = 1
				return partial
			else return nil, err end
		end
	})
end


_M.sourcet["default"] = _M.sourcet["until-closed"]

_M.source = _M.choose(_M.sourcet)

return _M
